from dataclasses import dataclass
from typing import Sequence

from .const import ControlEnum, OumanUnit

OumanValues = str | float | ControlEnum
"""Type alias for possible endpoint values: string, float, or a control enum."""


@dataclass(frozen=True)
class OumanEndpoint:
    """Base class for all Ouman device endpoints.

    Attributes:
        name: Human-readable name of the endpoint.
        unit: Unit of measurement, or None if not applicable.
        sensor_endpoint_id: The device API identifier for reading this endpoint.
    """

    name: str
    unit: OumanUnit | None
    sensor_endpoint_id: str

    def parse_value(self, value: str) -> OumanValues:
        return value


class ControllableEndpoint:
    """Base marker class for all writable endpoints."""

    pass


class NumberOumanEndpoint(OumanEndpoint):
    """Endpoint that returns numeric (float) values."""

    def parse_value(self, value: str) -> float:
        return float(value)


@dataclass(frozen=True)
class EnumControlOumanEndpoint(OumanEndpoint, ControllableEndpoint):
    """Controllable endpoint that accepts enum values.

    Attributes:
        control_endpoint_ids: API identifiers for writing to this endpoint.
        response_endpoint_ids: API identifiers expected in the response.
        enum_type: The enum class for valid values.
    """

    control_endpoint_ids: Sequence[str]
    response_endpoint_ids: Sequence[str]
    enum_type: type[ControlEnum]

    def parse_value(self, value: str) -> ControlEnum:
        return self.enum_type(value)


@dataclass(frozen=True)
class IntControlOumanEndpoint(NumberOumanEndpoint, ControllableEndpoint):
    """Controllable endpoint that accepts integer values.

    Attributes:
        control_endpoint_id: API identifier for writing to this endpoint.
        min_val: Minimum allowed value.
        max_val: Maximum allowed value.
    """

    control_endpoint_id: str
    min_val: int
    max_val: int


@dataclass(frozen=True)
class FloatControlOumanEndpoint(NumberOumanEndpoint, ControllableEndpoint):
    """Controllable endpoint that accepts float values.

    Attributes:
        control_endpoint_id: API identifier for writing to this endpoint.
        min_val: Minimum allowed value.
        max_val: Maximum allowed value.
    """

    control_endpoint_id: str
    min_val: float
    max_val: float



import asyncio
import logging
from datetime import datetime, timezone
from email.utils import formatdate
from typing import Iterable, Mapping, NamedTuple, Sequence

import aiohttp
from aiohttp import ClientSession

from .const import ControlEnum, HomeAwayControl, OperationMode
from .endpoint import (
    ControllableEndpoint,
    EnumControlOumanEndpoint,
    FloatControlOumanEndpoint,
    IntControlOumanEndpoint,
    OumanEndpoint,
    OumanValues,
)
from .exceptions import (
    OumanClientAuthenticationError,
    OumanClientCommunicationError,
    OumanClientError,
)
from .registry import (
    L1Endpoints,
    L1EndpointsWithRoomSensor,
    L2Endpoints,
    L2EndpointsWithRoomSensor,
    OumanRegistry,
    OumanRegistrySet,
    SystemEndpoints,
)

_LOGGER = logging.getLogger(__name__)


class _OumanResponse(NamedTuple):
    prefix: str
    values: Mapping[str, str]


class OumanEh800Client:
    """Client for communicating with an Ouman EH-800 heating controller.

    Args:
        session: An aiohttp ClientSession for making HTTP requests.
        address: The base URL of the device (e.g., "http://192.168.1.100").
        username: Username for authentication.
        password: Password for authentication.
    """

    def __init__(
        self, session: ClientSession, address: str, username: str, password: str
    ):
        self._session = session
        self._address = address
        self._username = username
        self._password = password

    @staticmethod
    def _parse_api_response(response_text: str) -> _OumanResponse:
        prefix, key_val_str = response_text.split("?", maxsplit=1)
        pairs = [p for p in key_val_str.split(";") if p.strip()]

        # Handle null byte at the end of the response
        if pairs and pairs[-1] == "\x00":
            pairs.pop()

        values_result = {}
        for pair in pairs:
            if "=" in pair:
                key, value = pair.split("=", 1)
                values_result[key.strip()] = value.strip()
            else:
                _LOGGER.warning(
                    "Skipping malformed key value pair in Ouman response: '%s'", pair
                )

        return _OumanResponse(
            prefix=prefix,
            values=values_result,
        )

    def _construct_request_url(self, path: str, params: Iterable[str]) -> str:
        request_url = f"{self._address}/{path}"
        # Append gmt string and equals symbol to match what the web UI does.
        # The requests work without this param as well, except when there are no other params.
        gmt_string_param = (
            formatdate(
                timeval=datetime.now(timezone.utc).timestamp(),
                localtime=False,
                usegmt=True,
            )
            + "="
        )
        params = list(params)
        params.append(gmt_string_param)
        request_url += "?" + ";".join(params)
        return request_url

    async def _request(self, path: str, params: Iterable[str]) -> _OumanResponse:
        request_url = self._construct_request_url(path, params)
        try:
            async with asyncio.timeout(10):
                async with self._session.get(request_url) as response:
                    response.raise_for_status()

                    response_text = await response.text()
                    _LOGGER.debug("Raw response from device: '%s'", response_text)

                    parsed_response = self._parse_api_response(response_text)
                    return parsed_response
        except asyncio.TimeoutError as err:
            raise OumanClientCommunicationError("Timeout connecting to device") from err
        except aiohttp.ClientResponseError as err:
            raise OumanClientCommunicationError(f"HTTP Error: {err.status}") from err
        except aiohttp.ClientError as err:
            raise OumanClientCommunicationError(f"Network error: {err}") from err

    async def login(self) -> None:
        """Authenticate with the Ouman EH-800 device.

        Raises:
            OumanClientAuthenticationError: If authentication fails.
            OumanClientError: If the response is unexpected.
        """
        response = await self._request(
            "login", [f"uid={self._username}", f"pwd={self._password}"]
        )

        if response.values.get("result") == "ok":
            _LOGGER.debug("Successful login")
        elif response.values.get("result") == "error":
            raise OumanClientAuthenticationError("Wrong username or password")
        else:
            raise OumanClientError(
                f"Unexpected response from login request: {response}"
            )

    async def _get_values(self, endpoint_ids: Sequence[str]) -> _OumanResponse:
        response = await self._request("request", endpoint_ids)
        for endpoint_id in endpoint_ids:
            if endpoint_id not in response.values:
                _LOGGER.warning(
                    "Requested endpoint ID '%s' not found in response", endpoint_id
                )
        return response

    async def get_values(
        self, registry_set: OumanRegistrySet
    ) -> dict[OumanEndpoint, OumanValues]:
        """Get all endpoint values for the specified registry set.

        Args:
            registry_set: The registry set containing endpoints to query.

        Returns:
            A dictionary mapping endpoints to their current values.
        """
        response = await self._get_values(registry_set.sensor_endpoint_ids)

        result = {}
        for key, value in response.values.items():
            endpoint = registry_set.get_endpoint_by_sensor_id(key)
            if not endpoint:
                _LOGGER.warning(f"Unexpected endpoint ID in response: '{key}'")
                continue
            parsed_value = endpoint.parse_value(value)
            result[endpoint] = parsed_value

        return result

    async def _update_values(
        self, key_value_params: Mapping[str, str]
    ) -> _OumanResponse:
        request_path = "update"
        params = [f"{key}={value}" for key, value in key_value_params.items()]
        try:
            response = await self._request(request_path, params)
        except OumanClientCommunicationError as err:
            if not isinstance(err.__cause__, aiohttp.ClientResponseError):
                raise
            if err.__cause__.status != 404:
                raise
            _LOGGER.debug("404 response from update request, logging in...")
            await self.login()
            response = await self._request(request_path, params)
        return response

    async def _set_int_endpoint(
        self, endpoint: IntControlOumanEndpoint, value: int
    ) -> int:
        if not (endpoint.min_val <= value <= endpoint.max_val):
            raise ValueError(
                f"Value for {endpoint.name} out of bounds [{endpoint.min_val},{endpoint.max_val}]: {value}"
            )
        params = {endpoint.control_endpoint_id: str(value)}
        result = await self._update_values(params)

        if not (result_value := result.values.get(endpoint.sensor_endpoint_id)):
            raise OumanClientError(
                f"Endpoint ID missing from set int endpoint response: {result}"
            )

        try:
            float_result = endpoint.parse_value(result_value)
        except ValueError as err:
            raise OumanClientError(
                f"API returned value cannot be parsed into a float: {result}"
            ) from err

        if float_result != value:
            raise OumanClientError(
                f"Returned float does not match set int value. Got {result_value}, expected {value}"
            )

        return int(float_result)

    async def _set_float_endpoint(
        self, endpoint: FloatControlOumanEndpoint, value: float
    ) -> float:
        """Sets an endpoint value for endpoints which accept floating
        point numbers. Values are rounded to the precision of one
        decimal."""
        if not (endpoint.min_val <= value <= endpoint.max_val):
            raise ValueError(
                f"Value for {endpoint.name} out of bounds [{endpoint.min_val},{endpoint.max_val}]: {value}"
            )
        rounded_value = round(value, 1)
        params = {endpoint.control_endpoint_id: str(round(value, 1))}
        result = await self._update_values(params)

        if not (result_value := result.values.get(endpoint.sensor_endpoint_id)):
            raise OumanClientError(
                f"Endpoint ID missing from set float endpoint response: {result}"
            )

        try:
            float_result = endpoint.parse_value(result_value)
        except ValueError as err:
            raise OumanClientError(
                f"API returned value cannot be parsed into a float: {result}"
            ) from err

        if float_result != rounded_value:
            raise OumanClientError(
                f"Returned float does not match set value. Got {result_value}, expected {value}"
            )

        return float_result

    async def _set_enum_endpoint(
        self, endpoint: EnumControlOumanEndpoint, value: ControlEnum
    ) -> ControlEnum:
        if not isinstance(value, endpoint.enum_type):
            raise TypeError(
                f"Unexpected type for {endpoint.name} value. Expected {endpoint.enum_type}, got {value}."
            )
        params = {endpoint_id: value for endpoint_id in endpoint.control_endpoint_ids}
        result = await self._update_values(params)

        for endpoint_id in endpoint.response_endpoint_ids:
            if not (result_value := result.values.get(endpoint_id)):
                raise OumanClientError(
                    f"Endpoint ID missing from set enum endpoint response: {result}"
                )

        if result_value != value:
            raise OumanClientError(
                f"Returned value does not match str enum value. Got '{result_value}', expected '{value}'"
            )
        try:
            enum_result = endpoint.parse_value(result_value)
        except ValueError as err:
            raise OumanClientError(
                f"API returned value cannot be parsed into an enum: {enum_result}"
            ) from err
        return enum_result

    async def set_endpoint_value(
        self,
        endpoint: ControllableEndpoint,
        value: OumanValues | int,
    ) -> OumanValues:
        """Set a value for a controllable endpoint.

        Args:
            endpoint: The controllable endpoint to set.
            value: The value to set for the endpoint.

        Returns:
            The confirmed value from the device.

        Raises:
            TypeError: If the endpoint is not controllable or value type is wrong.
            ValueError: If the value is out of bounds.
        """
        if not isinstance(endpoint, ControllableEndpoint):
            raise TypeError(f"Endpoint {endpoint} is not a controllable endpoint.")

        result: OumanValues
        if isinstance(endpoint, IntControlOumanEndpoint):
            if not isinstance(value, (int, float)):
                raise TypeError(
                    f"Value for {endpoint.name} must be numeric, got {type(value).__name__}"
                )
            if not value.is_integer():
                raise ValueError(
                    f"Value for {endpoint.name} must be an integer, got {value}"
                )
            result = await self._set_int_endpoint(endpoint, int(value))
        elif isinstance(endpoint, FloatControlOumanEndpoint):
            if not isinstance(value, (int, float)):
                raise TypeError(
                    f"Value for {endpoint.name} must be numeric, got {type(value).__name__}"
                )
            result = await self._set_float_endpoint(endpoint, value)
        elif isinstance(endpoint, EnumControlOumanEndpoint):
            if not isinstance(value, ControlEnum):
                raise TypeError(
                    f"Value for {endpoint.name} must be a ControlEnum, "
                    f"got {type(value).__name__}"
                )
            result = await self._set_enum_endpoint(endpoint, value)
        else:
            raise NotImplementedError(
                f"No control handler implemented for {type(endpoint).__name__}"
            )

        return result

    async def set_home_away(self, value: HomeAwayControl) -> HomeAwayControl:
        """Set the home/away mode for the system.

        Args:
            value: The home/away mode to set.

        Returns:
            The confirmed home/away mode from the device.
        """
        result = await self._set_enum_endpoint(SystemEndpoints.HOME_AWAY_MODE, value)
        if not isinstance(result, HomeAwayControl):
            raise TypeError(f"Unexpected return type: {type(result).__name__}")
        return result

    async def set_trend_sample_interval(self, value: int) -> int:
        """Set the trend sampling interval.

        Args:
            value: The sampling interval in seconds (30-21600).

        Returns:
            The confirmed sampling interval from the device.
        """
        result = await self._set_int_endpoint(
            SystemEndpoints.TREND_SAMPLE_INTERVAL, value
        )
        return result

    async def get_system_values(self) -> dict[OumanEndpoint, OumanValues]:
        """Get all values from the system endpoints registry.

        Returns:
            A dictionary mapping endpoints to their current values.
        """
        result = await self.get_values(OumanRegistrySet([SystemEndpoints]))
        return result

    async def set_l1_operation_mode(self, value: OperationMode) -> OperationMode:
        """Set the operation mode for L1 heating circuit.

        Args:
            value: The operation mode to set.

        Returns:
            The confirmed operation mode from the device.
        """
        result = await self._set_enum_endpoint(L1Endpoints.OPERATION_MODE, value)
        if not isinstance(result, OperationMode):
            raise TypeError(f"Unexpected return type: {type(result).__name__}")
        return result

    async def set_l1_valve_position_setpoint(self, position: int) -> int:
        """Set the valve position setpoint for L1 heating circuit.

        Args:
            position: The valve position percentage (0-100).

        Returns:
            The confirmed valve position from the device.
        """
        result = await self._set_int_endpoint(
            L1Endpoints.VALVE_POSITION_SETPOINT, position
        )
        return result

    async def set_l1_curve_minus_20_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at -20°C outdoor temp for L1.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L1Endpoints.CURVE_MINUS_20_TEMP, temperature
        )
        return result

    async def set_l1_curve_0_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at 0°C outdoor temp for L1.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L1Endpoints.CURVE_0_TEMP, temperature)
        return result

    async def set_l1_curve_20_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at +20°C outdoor temp for L1.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L1Endpoints.CURVE_20_TEMP, temperature)
        return result

    async def set_l1_temperature_drop(self, temperature: int) -> int:
        """Set the supply water setpoint for "temperature drop" operation mode on L1.

        This is the target supply water temperature used when the L1 heating
        circuit is set to "temperature drop" operation mode.

        Args:
            temperature: The supply water setpoint in Celsius (0-90).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L1Endpoints.TEMPERATURE_DROP, temperature)
        return result

    async def set_l1_big_temperature_drop(self, temperature: int) -> int:
        """Set the supply water setpoint for "big temperature drop" operation mode on L1.

        This is the target supply water temperature used when the L1 heating
        circuit is set to "big temperature drop" operation mode.

        Args:
            temperature: The supply water setpoint in Celsius (0-90).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L1Endpoints.BIG_TEMPERATURE_DROP, temperature
        )
        return result

    async def set_l1_water_out_minimum_temperature(self, temperature: int) -> int:
        """Set the minimum outgoing water temperature for L1 heating circuit.

        Args:
            temperature: The minimum temperature in Celsius (5-95).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L1Endpoints.WATER_OUT_MIN_TEMP, temperature
        )
        return result

    async def set_l1_water_out_maximum_temperature(self, temperature: int) -> int:
        """Set the maximum outgoing water temperature for L1 heating circuit.

        Args:
            temperature: The maximum temperature in Celsius (5-95).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L1Endpoints.WATER_OUT_MAX_TEMP, temperature
        )
        return result

    async def set_l1_room_temperature_fine_tuning(self, temperature: float) -> float:
        """Set the room temperature fine tuning offset for L1 (without room sensor).

        Args:
            temperature: The fine tuning offset in Celsius (-4.0 to 4.0).

        Returns:
            The confirmed offset from the device.
        """
        result = await self._set_float_endpoint(
            L1Endpoints.ROOM_TEMPERATURE_FINE_TUNING, temperature
        )
        return result

    async def set_l1_room_temperature_fine_tuning_with_sensor(
        self, temperature: float
    ) -> float:
        """Set the room temperature fine tuning offset for L1 (with room sensor).

        Args:
            temperature: The fine tuning offset in Celsius (-4.0 to 4.0).

        Returns:
            The confirmed offset from the device.
        """
        result = await self._set_float_endpoint(
            L1EndpointsWithRoomSensor.ROOM_TEMPERATURE_FINE_TUNING, temperature
        )
        return result

    async def get_l1_values(self) -> dict[OumanEndpoint, OumanValues]:
        """Get all values from the L1 heating circuit endpoints registry.

        Returns:
            A dictionary mapping endpoints to their current values.
        """
        result = await self.get_values(OumanRegistrySet([L1Endpoints]))
        return result

    async def set_l2_operation_mode(self, value: OperationMode) -> OperationMode:
        """Set the operation mode for L2 heating circuit.

        Args:
            value: The operation mode to set.

        Returns:
            The confirmed operation mode from the device.
        """
        result = await self._set_enum_endpoint(L2Endpoints.OPERATION_MODE, value)
        if not isinstance(result, OperationMode):
            raise TypeError(f"Unexpected return type: {type(result).__name__}")
        return result

    async def set_l2_valve_position_setpoint(self, position: int) -> int:
        """Set the valve position setpoint for L2 heating circuit.

        Args:
            position: The valve position percentage (0-100).

        Returns:
            The confirmed valve position from the device.
        """
        result = await self._set_int_endpoint(
            L2Endpoints.VALVE_POSITION_SETPOINT, position
        )
        return result

    async def set_l2_curve_minus_20_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at -20°C outdoor temp for L2.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L2Endpoints.CURVE_MINUS_20_TEMP, temperature
        )
        return result

    async def set_l2_curve_0_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at 0°C outdoor temp for L2.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L2Endpoints.CURVE_0_TEMP, temperature)
        return result

    async def set_l2_curve_20_temp(self, temperature: int) -> int:
        """Set the heating curve temperature at +20°C outdoor temp for L2.

        Args:
            temperature: The supply water temperature in Celsius (0-99).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L2Endpoints.CURVE_20_TEMP, temperature)
        return result

    async def set_l2_temperature_drop(self, temperature: int) -> int:
        """Set the supply water setpoint for "temperature drop" operation mode on L2.

        This is the target supply water temperature used when the L2 heating
        circuit is set to "temperature drop" operation mode.

        Args:
            temperature: The supply water setpoint in Celsius (0-90).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(L2Endpoints.TEMPERATURE_DROP, temperature)
        return result

    async def set_l2_big_temperature_drop(self, temperature: int) -> int:
        """Set the supply water setpoint for "big temperature drop" operation mode on L2.

        This is the target supply water temperature used when the L2 heating
        circuit is set to "big temperature drop" operation mode.

        Args:
            temperature: The supply water setpoint in Celsius (0-90).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L2Endpoints.BIG_TEMPERATURE_DROP, temperature
        )
        return result

    async def set_l2_water_out_minimum_temperature(self, temperature: int) -> int:
        """Set the minimum outgoing water temperature for L2 heating circuit.

        Args:
            temperature: The minimum temperature in Celsius (5-95).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L2Endpoints.WATER_OUT_MIN_TEMP, temperature
        )
        return result

    async def set_l2_water_out_maximum_temperature(self, temperature: int) -> int:
        """Set the maximum outgoing water temperature for L2 heating circuit.

        Args:
            temperature: The maximum temperature in Celsius (5-95).

        Returns:
            The confirmed temperature from the device.
        """
        result = await self._set_int_endpoint(
            L2Endpoints.WATER_OUT_MAX_TEMP, temperature
        )
        return result

    async def set_l2_room_temperature_fine_tuning(self, temperature: float) -> float:
        """Set the room temperature fine tuning offset for L2 (without room sensor).

        Args:
            temperature: The fine tuning offset in Celsius (-4.0 to 4.0).

        Returns:
            The confirmed offset from the device.
        """
        result = await self._set_float_endpoint(
            L2Endpoints.ROOM_TEMPERATURE_FINE_TUNING, temperature
        )
        return result

    async def set_l2_room_temperature_fine_tuning_with_sensor(
        self, temperature: float
    ) -> float:
        """Set the room temperature fine tuning offset for L2 (with room sensor).

        Args:
            temperature: The fine tuning offset in Celsius (-4.0 to 4.0).

        Returns:
            The confirmed offset from the device.
        """
        result = await self._set_float_endpoint(
            L2EndpointsWithRoomSensor.ROOM_TEMPERATURE_FINE_TUNING, temperature
        )
        return result

    async def get_l2_values(self) -> dict[OumanEndpoint, OumanValues]:
        """Get all values from the L2 heating circuit endpoints registry.

        Returns:
            A dictionary mapping endpoints to their current values.
        """
        result = await self.get_values(OumanRegistrySet([L2Endpoints]))
        return result

    async def get_is_l2_installed(self) -> bool:
        """Check if the L2 heating circuit is installed.

        Returns:
            True if L2 is installed, False otherwise.
        """
        endpoint_id = SystemEndpoints.L2_INSTALLED_STATUS.sensor_endpoint_id
        response = await self._get_values([endpoint_id])
        value = response.values.get(endpoint_id)
        if value is None:
            raise ValueError("Response value should be defined")

        # FIXME:
        # This is a temporary solution. It hasn't been tested what the
        # endpoint returns when L2 is installed, so now we
        # just assume that when the value differs from the falsy value,
        # it is means it is truthy and L2 is installed.
        return value != "0"

    async def _get_is_room_sensor_installed(self, endpoint_id: str) -> bool:
        response = await self._get_values([endpoint_id])
        value = response.values.get(endpoint_id)
        if value is None:
            raise ValueError("Response value should be defined")

        # FIXME:
        # This is a temporary solution. It hasn't been tested what the
        # endpoint returns when a room sensor is installed, so now we
        # just assume that when the value differs from the falsy value,
        # it is means it is truthy and a room sensor is installed.
        return value != "off"

    async def get_is_l1_room_sensor_installed(self) -> bool:
        """Check if a room sensor is installed for the L1 heating circuit.

        Returns:
            True if a room sensor is installed, False otherwise.
        """
        return await self._get_is_room_sensor_installed(
            L1Endpoints.ROOM_SENSOR_INSTALLED.sensor_endpoint_id
        )

    async def get_is_l2_room_sensor_installed(self) -> bool:
        """Check if a room sensor is installed for the L2 heating circuit.

        Returns:
            True if a room sensor is installed, False otherwise.
        """
        return await self._get_is_room_sensor_installed(
            L2Endpoints.ROOM_SENSOR_INSTALLED.sensor_endpoint_id
        )

    async def get_active_registries(self) -> OumanRegistrySet:
        """Get the list of active registries which contain the sets of
        endpoints that can currently be read and written to."""
        registries: list[type[OumanRegistry]] = [SystemEndpoints]
        if await self.get_is_l1_room_sensor_installed():
            registries.append(L1EndpointsWithRoomSensor)
        else:
            registries.append(L1Endpoints)
        if await self.get_is_l2_installed():
            if await self.get_is_l2_room_sensor_installed():
                registries.append(L2EndpointsWithRoomSensor)
            else:
                registries.append(L2Endpoints)
        return OumanRegistrySet(registries)

    async def get_alarms(self) -> Mapping[str, str]:
        """Get all active alarms from the device.

        Returns:
            A mapping of alarm identifiers to their values.
        """
        response = await self._request("alarms", [])
        return response.values

    async def logout(self) -> None:
        """Log out from the Ouman EH-800 device.

        Raises:
            OumanClientError: If the logout fails.
        """
        response = await self._request("logout", [])
        if response.values.get("result") != "ok":
            raise OumanClientError(
                f"Unexpected response from logout request: {response}"
            )



class OumanClientError(Exception):
    """Base exception for all Ouman client errors."""

    pass


class OumanClientAuthenticationError(OumanClientError):
    """Raised when authentication with the device fails."""

    pass


class OumanClientCommunicationError(OumanClientError):
    """Raised when communication with the device fails."""

    pass



from enum import StrEnum


class HomeAwayControl(StrEnum):
    """Control enum for home/away mode setting."""

    HOME = "0"
    AWAY = "1"
    OFF = "2"


class OperationMode(StrEnum):
    """Control enum for heating circuit operation mode."""

    AUTOMATIC = "0"
    TEMPERATURE_DROP = "1"
    BIG_TEMPERATURE_DROP = "2"
    NORMAL_TEMPERATURE = "3"
    SHUTDOWN = "5"
    MANUAL_VALVE_CONTROL = "6"


ControlEnum = HomeAwayControl | OperationMode


class OumanUnit(StrEnum):
    """Unit of measurement for endpoint values."""

    CELSIUS = "°C"
    SECOND = "s"
    PERCENT = "%"



from .client import OumanEh800Client
from .const import HomeAwayControl, OperationMode, OumanUnit
from .endpoint import (
    ControllableEndpoint,
    EnumControlOumanEndpoint,
    FloatControlOumanEndpoint,
    IntControlOumanEndpoint,
    NumberOumanEndpoint,
    OumanEndpoint,
    OumanValues,
)
from .exceptions import (
    OumanClientAuthenticationError,
    OumanClientCommunicationError,
    OumanClientError,
)
from .registry import (
    L1Endpoints,
    L1EndpointsWithRoomSensor,
    L2Endpoints,
    L2EndpointsWithRoomSensor,
    OumanRegistry,
    OumanRegistrySet,
    SystemEndpoints,
)

__all__ = [
    # Client
    "OumanEh800Client",
    # Enums
    "HomeAwayControl",
    "OperationMode",
    "OumanUnit",
    # Registry
    "OumanRegistry",
    "OumanRegistrySet",
    "L1Endpoints",
    "L1EndpointsWithRoomSensor",
    "L2Endpoints",
    "L2EndpointsWithRoomSensor",
    "SystemEndpoints",
    # Endpoint types
    "OumanEndpoint",
    "OumanValues",
    "ControllableEndpoint",
    "NumberOumanEndpoint",
    "IntControlOumanEndpoint",
    "FloatControlOumanEndpoint",
    "EnumControlOumanEndpoint",
    # Exceptions
    "OumanClientError",
    "OumanClientAuthenticationError",
    "OumanClientCommunicationError",
]



from dataclasses import dataclass
from functools import cached_property
from typing import Generator, Mapping, Sequence

from .const import HomeAwayControl, OperationMode, OumanUnit
from .endpoint import (
    EnumControlOumanEndpoint,
    FloatControlOumanEndpoint,
    IntControlOumanEndpoint,
    NumberOumanEndpoint,
    OumanEndpoint,
)


class OumanRegistry:
    """Base class for endpoint registry definitions.

    Subclasses define endpoints as class attributes. Child registries
    can override parent endpoints by redefining them with the same name.
    """

    @classmethod
    def iterate_endpoints(cls) -> Generator[OumanEndpoint, None, None]:
        """Iterate over all OumanEndpoints in this class and its parents.

        Subclass definitions override parent class definitions.
        """
        seen_keys = set()
        for base in cls.mro():
            for key, value in base.__dict__.items():
                # Avoid yielding the same endpoint twice if overridden
                if key not in seen_keys and isinstance(value, OumanEndpoint):
                    seen_keys.add(key)
                    yield value


@dataclass
class OumanRegistrySet:
    """A collection of registries for querying endpoint values.

    Use this to group registries when calling client.get_values().
    Validates that registries don't have conflicting endpoint IDs.
    """

    registries: Sequence[type[OumanRegistry]]

    def __post_init__(self) -> None:
        if len(self.registries) > len(set(self.registries)):
            raise ValueError("Multiple of the same registry passed")

        if len(self.endpoints) > len(self._sensor_id_endpoint_map):
            raise ValueError("Conflicting endpoint IDs across registries")

    @cached_property
    def endpoints(self) -> Sequence[OumanEndpoint]:
        """All the endpoints in the registry set."""
        return [
            endpoint
            for registry in self.registries
            for endpoint in registry.iterate_endpoints()
        ]

    @cached_property
    def _sensor_id_endpoint_map(self) -> Mapping[str, OumanEndpoint]:
        return {endpoint.sensor_endpoint_id: endpoint for endpoint in self.endpoints}

    @cached_property
    def sensor_endpoint_ids(self) -> Sequence[str]:
        return [endpoint.sensor_endpoint_id for endpoint in self.endpoints]

    def get_endpoint_by_sensor_id(self, id: str) -> OumanEndpoint | None:
        return self._sensor_id_endpoint_map.get(id)


class SystemEndpoints(OumanRegistry):
    """System-wide endpoints for the Ouman EH-800 device."""

    TREND_SAMPLE_INTERVAL = IntControlOumanEndpoint(
        name="trend_sampling_interval",
        unit=OumanUnit.SECOND,
        sensor_endpoint_id="S_26_85",
        control_endpoint_id="@_S_26_85",
        min_val=30,
        max_val=21600,
    )

    HOME_AWAY_MODE = EnumControlOumanEndpoint(
        name="home_away_mode",
        unit=None,
        sensor_endpoint_id="S_135_85",
        control_endpoint_ids=("S_135_85", "S_222_85"),
        response_endpoint_ids=("S_222_85",),
        enum_type=HomeAwayControl,
    )

    OUTSIDE_TEMPERATURE = NumberOumanEndpoint(
        name="outside_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_227_85",
    )

    RELAY_CONFIGURATION_TYPE = OumanEndpoint(
        name="relay_configuration_type",
        unit=None,
        sensor_endpoint_id="S_1002_85",
    )

    RELAY_STATUS_TEXT = OumanEndpoint(
        name="relay_status_text",
        unit=None,
        sensor_endpoint_id="S_1004_85",
    )

    L2_INSTALLED_STATUS = OumanEndpoint(
        name="l2_installed_status",
        unit=None,
        sensor_endpoint_id="S_140_85",
    )


class L1Endpoints(OumanRegistry):
    """Endpoints for the L1 (primary) heating circuit."""

    OPERATION_MODE = EnumControlOumanEndpoint(
        name="l1_operation_mode",
        unit=None,
        sensor_endpoint_id="S_59_85",
        control_endpoint_ids=("S_59_85",),
        response_endpoint_ids=("S_59_85",),
        enum_type=OperationMode,
    )

    VALVE_POSITION_SETPOINT = IntControlOumanEndpoint(
        name="l1_valve_position_setpoint",
        unit=OumanUnit.PERCENT,
        sensor_endpoint_id="S_92_85",
        control_endpoint_id="S_92_85",
        min_val=0,
        max_val=100,
    )

    CURVE_MINUS_20_TEMP = IntControlOumanEndpoint(
        name="l1_curve_minus_20_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_61_85",
        control_endpoint_id="@_S_61_85",
        min_val=0,
        max_val=99,
    )

    CURVE_0_TEMP = IntControlOumanEndpoint(
        name="l1_curve_0_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_63_85",
        control_endpoint_id="@_S_63_85",
        min_val=0,
        max_val=99,
    )

    CURVE_20_TEMP = IntControlOumanEndpoint(
        name="l1_curve_20_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_65_85",
        control_endpoint_id="@_S_65_85",
        min_val=0,
        max_val=99,
    )

    TEMPERATURE_DROP = IntControlOumanEndpoint(
        name="l1_temperature_drop",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_89_85",
        control_endpoint_id="@_S_89_85",
        min_val=0,
        max_val=90,
    )

    BIG_TEMPERATURE_DROP = IntControlOumanEndpoint(
        name="l1_big_temperature_drop",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_90_85",
        control_endpoint_id="@_S_90_85",
        min_val=0,
        max_val=90,
    )

    WATER_OUT_MIN_TEMP = IntControlOumanEndpoint(
        name="l1_water_out_minimum_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_54_85",
        control_endpoint_id="@_S_54_85",
        min_val=5,
        max_val=95,
    )

    WATER_OUT_MAX_TEMP = IntControlOumanEndpoint(
        name="l1_water_out_maximum_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_55_85",
        control_endpoint_id="@_S_55_85",
        min_val=5,
        max_val=95,
    )

    ROOM_TEMPERATURE_FINE_TUNING = FloatControlOumanEndpoint(
        name="l1_room_temperature_fine_tuning",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_134_85",
        control_endpoint_id="@_S_134_85",
        min_val=-4.0,
        max_val=4.0,
    )

    HEATING_SHUTDOWN_STATUS = OumanEndpoint(
        name="l1_heating_shutdown_status",
        unit=None,
        sensor_endpoint_id="S_0_0",
    )

    TEMPERATURE_LEVEL_STATUS_TEXT = OumanEndpoint(
        name="l1_temperature_level_status_text",
        unit=None,
        sensor_endpoint_id="S_1000_0",
    )

    CIRCUIT_NAME = OumanEndpoint(
        name="l1_circuit_name",
        unit=None,
        sensor_endpoint_id="S_131_85",
    )

    SUPPLY_WATER_TEMPERATURE = NumberOumanEndpoint(
        name="l1_supply_water_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_259_85",
    )

    VALVE_POSITION = NumberOumanEndpoint(
        name="l1_valve_position",
        unit=OumanUnit.PERCENT,
        sensor_endpoint_id="S_272_85",
    )

    SUPPLY_WATER_TEMPERATURE_SETPOINT = NumberOumanEndpoint(
        name="l1_supply_water_temperature_setpoint",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_275_85",
    )

    # TODO: make this into an enum endpoint when we know what other
    # values are possible besides "off"
    ROOM_SENSOR_INSTALLED = OumanEndpoint(
        name="l1_room_sensor_installed",
        unit=None,
        sensor_endpoint_id="S_261_111",
    )


class L1EndpointsWithRoomSensor(L1Endpoints):
    """Endpoints for L1 heating circuit with a room sensor installed.

    Note: The endpoints in this registry have not been verified.

    Extends L1Endpoints with additional room sensor endpoints and
    overrides ROOM_TEMPERATURE_FINE_TUNING with the correct control endpoint.
    """

    ROOM_SENSOR_POTENTIOMETER = NumberOumanEndpoint(
        name="l1_room_sensor_potentiometer",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_274_85",
    )

    ROOM_TEMPERATURE = NumberOumanEndpoint(
        name="l1_room_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_261_85",
    )

    ROOM_TEMPERATURE_SETPOINT = NumberOumanEndpoint(
        name="l1_room_temperature_setpoint",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_278_85",
    )

    # NOTE: This overrides the one in L1Endpoints since the control
    # endpoint is different when a room sensor is installed.
    ROOM_TEMPERATURE_FINE_TUNING = FloatControlOumanEndpoint(
        name="l1_room_temperature_fine_tuning",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_102_85",
        control_endpoint_id="@_S_102_85",
        min_val=-4.0,
        max_val=4.0,
    )


class L2Endpoints(OumanRegistry):
    """Endpoints for the L2 (secondary) heating circuit.

    Note: The endpoints in this registry have not been verified.
    """

    OPERATION_MODE = EnumControlOumanEndpoint(
        name="l2_operation_mode",
        unit=None,
        sensor_endpoint_id="S_146_85",
        control_endpoint_ids=("S_146_85",),
        response_endpoint_ids=("S_146_85",),
        enum_type=OperationMode,
    )

    VALVE_POSITION_SETPOINT = IntControlOumanEndpoint(
        name="l2_valve_position_setpoint",
        unit=OumanUnit.PERCENT,
        sensor_endpoint_id="S_179_85",
        control_endpoint_id="S_179_85",
        min_val=0,
        max_val=100,
    )

    CURVE_MINUS_20_TEMP = IntControlOumanEndpoint(
        name="l2_curve_minus_20_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_148_85",
        control_endpoint_id="@_S_148_85",
        min_val=0,
        max_val=99,
    )

    CURVE_0_TEMP = IntControlOumanEndpoint(
        name="l2_curve_0_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_150_85",
        control_endpoint_id="@_S_150_85",
        min_val=0,
        max_val=99,
    )

    CURVE_20_TEMP = IntControlOumanEndpoint(
        name="l2_curve_20_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_152_85",
        control_endpoint_id="@_S_152_85",
        min_val=0,
        max_val=99,
    )

    TEMPERATURE_DROP = IntControlOumanEndpoint(
        name="l2_temperature_drop",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_176_85",
        control_endpoint_id="@_S_176_85",
        min_val=0,
        max_val=90,
    )

    BIG_TEMPERATURE_DROP = IntControlOumanEndpoint(
        name="l2_big_temperature_drop",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_177_85",
        control_endpoint_id="@_S_177_85",
        min_val=0,
        max_val=90,
    )

    WATER_OUT_MIN_TEMP = IntControlOumanEndpoint(
        name="l2_water_out_minimum_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_141_85",
        control_endpoint_id="@_S_141_85",
        min_val=5,
        max_val=95,
    )

    WATER_OUT_MAX_TEMP = IntControlOumanEndpoint(
        name="l2_water_out_maximum_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_142_85",
        control_endpoint_id="@_S_142_85",
        min_val=5,
        max_val=95,
    )

    ROOM_TEMPERATURE_FINE_TUNING = FloatControlOumanEndpoint(
        name="l2_room_temperature_fine_tuning",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_221_85",
        control_endpoint_id="@_S_221_85",
        min_val=-4.0,
        max_val=4.0,
    )

    TEMPERATURE_LEVEL_STATUS_TEXT = OumanEndpoint(
        name="l2_temperature_level_status_text",
        unit=None,
        sensor_endpoint_id="S_1001_0",
    )

    CIRCUIT_NAME = OumanEndpoint(
        name="l2_circuit_name",
        unit=None,
        sensor_endpoint_id="S_218_85",
    )

    SUPPLY_WATER_TEMPERATURE = NumberOumanEndpoint(
        name="l2_supply_water_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_293_85",
    )

    SUPPLY_WATER_TEMPERATURE_SETPOINT = NumberOumanEndpoint(
        name="l2_supply_water_temperature_setpoint",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_310_85",
    )

    # TODO: make this into an enum endpoint when we know what other
    # values are possible besides "off"
    ROOM_SENSOR_INSTALLED = OumanEndpoint(
        name="l2_room_sensor_installed",
        unit=None,
        sensor_endpoint_id="S_295_111",
    )


class L2EndpointsWithRoomSensor(L2Endpoints):
    """Endpoints for L2 heating circuit with a room sensor installed.

    Note: The endpoints in this registry have not been verified.

    Extends L2Endpoints with additional room sensor endpoints and
    overrides ROOM_TEMPERATURE_FINE_TUNING with the correct control endpoint.
    """

    ROOM_SENSOR_POTENTIOMETER = NumberOumanEndpoint(
        name="l2_room_sensor_potentiometer",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_307_85",
    )

    ROOM_TEMPERATURE = NumberOumanEndpoint(
        name="l2_room_temperature",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_295_85",
    )

    ROOM_TEMPERATURE_SETPOINT = NumberOumanEndpoint(
        name="l2_room_temperature_setpoint",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_313_85",
    )

    # NOTE: This overrides the one in L2Endpoints since the control
    # endpoint is different when a room sensor is installed.
    ROOM_TEMPERATURE_FINE_TUNING = FloatControlOumanEndpoint(
        name="l2_room_temperature_fine_tuning",
        unit=OumanUnit.CELSIUS,
        sensor_endpoint_id="S_189_85",
        control_endpoint_id="@_S_189_85",
        min_val=-4.0,
        max_val=4.0,
    )
